# -*- coding: utf-8 -*-
"""
This module defines base classes for desktop notifications. All platform implementations
must inherit from :class:`DesktopNotifierBase`.
"""

# system imports
from enum import Enum
from typing import Optional, Dict, Callable, Union


class NotificationLevel(Enum):
    """Enumeration of notification levels

    The interpretation and visuals will depend on the platform.

    :cvar Critical: For critical errors.
    :cvar Normal: Default platform notification level.
    :cvar Low: Low priority notification.
    """

    Critical = "critical"
    Normal = "normal"
    Low = "low"


class Notification:
    """A desktop notification

    :param title: Notification title.
    :param message: Notification message.
    :param urgency: Notification level: low, normal or critical. This is ignored by some
        implementations.
    :param icon: Path to an icon to use for the notification, typically the app icon.
        This is ignored by some implementations, e.g., on macOS where the icon of the
        app bundle is always used.
    :param action: Handler to call when the notification is clicked. This is ignored by
        some implementations.
    :param buttons: A dictionary with button names to show in the notification and
        handler to call when the respective button is clicked. This is ignored by some
        implementations.

    :ivar identifier: An identifier which gets assigned to the notification after it is
        sent. This may be a str or int, depending on the type of identifier used by the
        platform.
    """

    identifier: Union[str, int, None]

    def __init__(
        self,
        title: str,
        message: str,
        urgency: NotificationLevel = NotificationLevel.Normal,
        icon: Optional[str] = None,
        action: Optional[Callable] = None,
        buttons: Optional[Dict[str, Callable]] = None,
    ) -> None:

        self.title = title
        self.message = message
        self.urgency = urgency
        self.icon = icon
        self.action = action
        self.buttons = buttons or dict()
        self.identifier = None

    def __repr__(self):
        return f"<{self.__class__.__name__}(title='{self.title}', message='{self.message}')>"


class DesktopNotifierBase:
    """Base class for desktop notifier implementations

    Notification levels CRITICAL, NORMAL and LOW may be used by some implementations to
    determine how a notification is displayed.

    :param app_name: Name to identify the application in the notification center.
    :param notification_limit: Maximum number of notifications to keep in the system's
        notification center.
    """

    app_name: str
    notification_limit: int
    current_notifications: Dict[int, Notification]

    def __init__(self, app_name: str = "", notification_limit: int = 5) -> None:
        self.app_name = app_name
        self.notification_limit = notification_limit
        self.current_notifications = dict()
        self._current_nid = 0

    def send(self, notification: Notification) -> None:
        """
        Sends a desktop notification. Some arguments may be ignored, depending on the
        implementation.

        :param notification: Notification to send.
        """
        pass

    def _next_nid(self) -> int:
        self._current_nid += 1
        self._current_nid %= self.notification_limit
        return self._current_nid



# -*- coding: utf-8 -*-
"""
UNUserNotificationCenter backend for macOS.

* Introduced in macOS 10.14.
* Cross-platform with iOS and iPadOS.
* Only available from signed app bundles if called from the main executable or from a
  signed Python framework (for example from python.org).
* Requires a running CFRunLoop to invoke callbacks.

"""

# system imports
import uuid
import logging
from concurrent.futures import Future, wait
from typing import Optional, Dict, Tuple

# external imports
from rubicon.objc import NSObject, ObjCClass, objc_method, py_from_ns  # type: ignore
from rubicon.objc.runtime import load_library, objc_id, objc_block  # type: ignore

# local imports
from .base import Notification, DesktopNotifierBase


__all__ = ["CocoaNotificationCenter"]

logger = logging.getLogger(__name__)

foundation = load_library("Foundation")
uns = load_library("UserNotifications")

UNUserNotificationCenter = ObjCClass("UNUserNotificationCenter")
UNMutableNotificationContent = ObjCClass("UNMutableNotificationContent")
UNNotificationRequest = ObjCClass("UNNotificationRequest")
UNNotificationAction = ObjCClass("UNNotificationAction")
UNNotificationCategory = ObjCClass("UNNotificationCategory")

NSSet = ObjCClass("NSSet")

UNNotificationDefaultActionIdentifier = (
    "com.apple.UNNotificationDefaultActionIdentifier"
)
UNNotificationDismissActionIdentifier = (
    "com.apple.UNNotificationDismissActionIdentifier"
)

UNAuthorizationOptionBadge = 1 << 0
UNAuthorizationOptionSound = 1 << 1
UNAuthorizationOptionAlert = 1 << 2

UNNotificationActionOptionForeground = 1 << 2

UNNotificationCategoryOptionNone = 0

UNAuthorizationStatusAuthorized = 2
UNAuthorizationStatusProvisional = 3
UNAuthorizationStatusEphemeral = 4


class NotificationCenterDelegate(NSObject):  # type: ignore
    """Delegate to handle user interactions with notifications"""

    @objc_method
    def userNotificationCenter_didReceiveNotificationResponse_withCompletionHandler_(
        self, center, response, completion_handler: objc_block
    ) -> None:

        # Get the notification which was clicked from the platform ID.
        internal_nid = py_from_ns(
            response.notification.request.content.userInfo["internal_nid"]
        )
        notification = self.interface.current_notifications[internal_nid]

        # Get and call the callback which corresponds to the user interaction.
        if response.actionIdentifier == UNNotificationDefaultActionIdentifier:

            callback = notification.action

            if callback:
                callback()

        elif response.actionIdentifier != UNNotificationDismissActionIdentifier:

            action_id_str = py_from_ns(response.actionIdentifier)

            callback = notification.buttons.get(action_id_str)

            if callback:
                callback()

        completion_handler()


class CocoaNotificationCenter(DesktopNotifierBase):
    """UNUserNotificationCenter backend for macOS

    Can be used with macOS Catalina and newer. Both app name and bundle identifier
    will be ignored. The notification center automatically uses the values provided
    by the app bundle.

    :param app_name: The name of the app. Does not have any effect because the app
        name is automatically determined from the bundle or framework.
    """

    _notification_categories: Dict[Tuple[str, ...], str]

    def __init__(self, app_name: str) -> None:
        super().__init__(app_name)
        self.nc = UNUserNotificationCenter.currentNotificationCenter()
        self.nc_delegate = NotificationCenterDelegate.alloc().init()
        self.nc_delegate.interface = self

        if self.nc:
            self.nc.delegate = self.nc_delegate
        else:
            logger.warning("No notification center available")

        self._did_request_authorisation = False
        self._notification_categories = {}

    def _request_authorisation(self) -> None:
        def on_auth_completed(granted: bool, error: objc_id) -> None:
            if granted:
                logger.debug("Authorisation granted")
            else:
                logger.debug("Authorisation denied")

            if error:
                error = py_from_ns(error)
                logger.warning("Authorisation error %s", str(error))

        self.nc.requestAuthorizationWithOptions(
            UNAuthorizationOptionAlert
            | UNAuthorizationOptionSound
            | UNAuthorizationOptionBadge,
            completionHandler=on_auth_completed,
        )

        self._did_request_authorisation = True

    @property
    def _has_authorisation(self) -> bool:

        # get existing notification categories

        future: Future = Future()

        def handler(settings: objc_id) -> None:
            settings = py_from_ns(settings)
            settings.retain()
            future.set_result(settings)

        self.nc.getNotificationSettingsWithCompletionHandler(handler)

        wait([future])
        settings = future.result()

        authorized = settings.authorizationStatus in (
            UNAuthorizationStatusAuthorized,
            UNAuthorizationStatusProvisional,
            UNAuthorizationStatusEphemeral,
        )

        settings.release()

        return authorized

    def send(self, notification: Notification) -> None:
        """
        Sends a notification.

        :param notification: Notification to send.
        """

        if not self.nc:
            return

        if not self._did_request_authorisation:
            self._request_authorisation()

        if not self._has_authorisation:
            return

        # Get an internal ID for the notifications. This will recycle an old ID if
        # we are above the max number of notifications.
        internal_nid = self._next_nid()

        # Get the old notification to replace, if any.
        notification_to_replace = self.current_notifications.get(internal_nid)

        if notification_to_replace:
            platform_nid = notification_to_replace.identifier
        else:
            platform_nid = str(uuid.uuid4())

        # Set up buttons for notification. On macOS, we need need to register a new
        # notification category for every unique set of buttons.
        button_names = tuple(notification.buttons.keys())
        category_id = self._category_id_for_button_names(button_names)

        # Create the native notification + notification request.
        content = UNMutableNotificationContent.alloc().init()
        content.title = notification.title
        content.body = notification.message
        content.categoryIdentifier = category_id
        content.userInfo = {"internal_nid": internal_nid}

        notification_request = UNNotificationRequest.requestWithIdentifier(
            platform_nid, content=content, trigger=None
        )

        # Post the notification.
        self.nc.addNotificationRequest(notification_request, withCompletionHandler=None)

        # Store the notification for future replacement and to keep track of
        # user-supplied callbacks.
        notification.identifier = platform_nid
        self.current_notifications[internal_nid] = notification

    def _category_id_for_button_names(
        self, button_names: Tuple[str, ...]
    ) -> Optional[str]:
        """
        Creates and registers a new notification category with the given buttons
        or retrieves an existing one.
        """

        if not button_names:
            return None

        try:
            return self._notification_categories[button_names]
        except KeyError:
            actions = []

            for name in button_names:
                action = UNNotificationAction.actionWithIdentifier(
                    name, title=name, options=UNNotificationActionOptionForeground
                )
                actions.append(action)

            # get existing notification categories

            future: Future = Future()

            def handler(categories: objc_id) -> None:
                categories = py_from_ns(categories)
                categories.retain()
                future.set_result(categories)

            self.nc.getNotificationCategoriesWithCompletionHandler(handler)

            wait([future])
            categories = future.result()

            # add category for new set of buttons

            category_id = str(uuid.uuid4())
            new_categories = categories.setByAddingObject(
                UNNotificationCategory.categoryWithIdentifier(
                    category_id,
                    actions=actions,
                    intentIdentifiers=[],
                    options=UNNotificationCategoryOptionNone,
                )
            )
            self.nc.setNotificationCategories(new_categories)
            self._notification_categories[button_names] = category_id

            categories.release()

            return category_id



# -*- coding: utf-8 -*-
"""
NSUserNotificationCenter backend for macOS.

* Should be used for macOS 10.13 and earlier.
* Deprecated but still available in macOS 11.0.
* Requires a running CFRunLoop to invoke callbacks.

"""

# system imports
import uuid
import platform
import logging

# external imports
from rubicon.objc import NSObject, ObjCClass, objc_method, py_from_ns  # type: ignore
from rubicon.objc.runtime import load_library  # type: ignore

# local imports
from .base import Notification, DesktopNotifierBase


__all__ = ["CocoaNotificationCenterLegacy"]

logger = logging.getLogger(__name__)
macos_version, *_ = platform.mac_ver()

foundation = load_library("Foundation")

NSUserNotification = ObjCClass("NSUserNotification")
NSUserNotificationCenter = ObjCClass("NSUserNotificationCenter")
NSDate = ObjCClass("NSDate")

NSUserNotificationActivationTypeContentsClicked = 1
NSUserNotificationActivationTypeActionButtonClicked = 2
NSUserNotificationActivationTypeAdditionalActionClicked = 4


class NotificationCenterDelegate(NSObject):  # type: ignore
    """Delegate to handle user interactions with notifications"""

    @objc_method
    def userNotificationCenter_didActivateNotification_(
        self, center, notification
    ) -> None:

        internal_nid = py_from_ns(notification.userInfo["internal_nid"])
        notification_info = self.interface.current_notifications[internal_nid]

        if (
            notification.activationType
            == NSUserNotificationActivationTypeContentsClicked
        ):

            if notification_info.action:
                notification_info.action()

        elif (
            notification.activationType
            == NSUserNotificationActivationTypeActionButtonClicked
        ):

            button_title = py_from_ns(notification.actionButtonTitle)
            callback = notification_info.buttons.get(button_title)

            if callback:
                callback()


class CocoaNotificationCenterLegacy(DesktopNotifierBase):
    """NSUserNotificationCenter backend for macOS

    Should be used for macOS High Sierra and earlier. Supports only a single button per
    notification. Both app name and bundle identifier will be ignored. The notification
    center automatically uses the values provided by the app bundle or the Python
    framework.

    :param app_name: The name of the app. Does not have any effect because the app
        name is automatically determined from the bundle or framework.
    """

    def __init__(self, app_name: str) -> None:
        super().__init__(app_name)

        self.nc = NSUserNotificationCenter.defaultUserNotificationCenter
        self.nc_delegate = NotificationCenterDelegate.alloc().init()
        self.nc_delegate.interface = self

        if self.nc:
            self.nc.delegate = self.nc_delegate
        else:
            logger.warning("No notification center available")

    def send(self, notification: Notification) -> None:
        """
        Sends a notification.

        :param notification: Notification to send.
        """

        if not self.nc:
            return

        internal_nid = self._next_nid()
        notification_to_replace = self.current_notifications.get(internal_nid)

        if notification_to_replace:
            platform_nid = notification_to_replace.identifier
        else:
            platform_nid = str(uuid.uuid4())

        n = NSUserNotification.alloc().init()
        n.title = notification.title
        n.informativeText = notification.message
        n.identifier = platform_nid
        n.userInfo = {"internal_nid": internal_nid}
        n.deliveryDate = NSDate.dateWithTimeInterval(0, sinceDate=NSDate.date())

        if notification.buttons:
            if len(notification.buttons) > 1:
                logger.debug(
                    "NSUserNotificationCenter: only a single button is supported"
                )
            n.hasActionButton = True
            n.actionButtonTitle = list(notification.buttons.keys())[0]

        self.nc.scheduleNotification(n)

        notification.identifier = platform_nid
        self.current_notifications[internal_nid] = notification



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages


requirements = [
    "dbus-next;sys_platform=='linux'",
    "packaging",
    "rubicon-objc;sys_platform=='darwin'",
]

dev_requires = [
    "black",
    "bump2version",
    "flake8",
    "mypy",
    "pre-commit",
    "pytest",
    "pytest-cov",
]

docs_require = [
    "sphinx",
    "m2r2",
    "sphinx-autoapi",
    "sphinx_rtd_theme",
]

setup(
    author="Sam Schott",
    author_email="sam.schott@outlook.com",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    description="Python library for cross-platform desktop notifications",
    setup_requires=["wheel"],
    install_requires=requirements,
    extras_require={
        "dev": dev_requires,
        "docs": docs_require,
    },
    license="MIT license",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords="desktop-notifier",
    name="desktop-notifier",
    packages=find_packages("src"),
    package_dir={"": "src"},
    url="https://github.com/samschott/desktop-notifier",
    version="1.1.0",
    zip_safe=False,
)



# -*- coding: utf-8 -*-
"""
This module handles desktop notifications and supports multiple backends, depending on
the platform.
"""

# system imports
import platform
from threading import Lock
from typing import Type, Optional, Dict, Callable

# external imports
from packaging.version import Version

# local imports
from .base import NotificationLevel, Notification, DesktopNotifierBase


Impl: Type[DesktopNotifierBase]

if platform.system() == "Darwin":

    macos_version, *_ = platform.mac_ver()

    if Version(macos_version) >= Version("10.14.0"):
        from .macos import CocoaNotificationCenter as Impl
    else:
        from .macos_legacy import CocoaNotificationCenterLegacy as Impl

elif platform.system() == "Linux":
    from .dbus import DBusDesktopNotifier as Impl

else:
    from .base import DesktopNotifierBase as Impl


__all__ = [
    "Notification",
    "NotificationLevel",
    "DesktopNotifier",
]


class DesktopNotifier:
    """Cross-platform desktop notification emitter

    Uses different backends depending on the platform version and available services.
    All implementations will dispatch notifications without an event loop but will
    require a running event loop to execute callbacks when the end user interacts with a
    notification. On Linux, a asyncio event loop is required. On macOS, a CFRunLoop *in
    the main thread* is required. Packages such as :mod:`rubicon.objc` can be used to
    integrate asyncio with a CFRunLoop.

    :param app_name: Name to identify the application in the notification center. On
        Linux, this should correspond to the application name in a desktop entry. On
        macOS, this argument is ignored and the app is identified by the bundle ID of
        the sending program (e.g., Python).
    :param notification_limit: Maximum number of notifications to keep in the system's
        notification center. This may be ignored by some implementations.
    """

    def __init__(self, app_name: str = "Python", notification_limit: int = 5) -> None:
        self._lock = Lock()
        self._impl = Impl(app_name, notification_limit)

    def send(
        self,
        title: str,
        message: str,
        urgency: NotificationLevel = NotificationLevel.Normal,
        icon: Optional[str] = None,
        action: Optional[Callable] = None,
        buttons: Optional[Dict[str, Callable]] = None,
    ) -> None:
        """
        Sends a desktop notification. Some arguments may be ignored, depending on the
        backend.

        :param title: Notification title.
        :param message: Notification message.
        :param urgency: Notification level: low, normal or critical. This may be
            interpreted differently but some implementations, for instance causing the
            notification to remain visible for longer, or may be ignored.
        :param icon: Path to an icon to use for the notification, typically the app
            icon. This is ignored by some implementations, e.g., on macOS where the icon
            of the app bundle is always used.
        :param action: Handler to call when the notification is clicked. This is ignored
            by some implementations.
        :param buttons: A dictionary with button names and callbacks to show in the
            notification. This is ignored by some implementations.
        """
        notification = Notification(title, message, urgency, icon, action, buttons)

        with self._lock:
            self._impl.send(notification)
